from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from measurement_stats import angle
from measurement_stats import mean
from measurement_stats import ops
from measurement_stats import value
from measurement_stats import value2D
from measurement_stats import values
from measurement_stats import density



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import math

import numpy as np


def uncertainty_estimate(values):
    """
    Creates an estimate for the uncertainty to apply to each value base
    on the separation of the entries that make up the distribution. This is
    used in cases where no uncertainty was described in the measurements,
    but the measurement population is fairly large, so that the
    quantization of the measurements themselves indicate the uncertainty
    implicitly.

    :param values: An iterable containing entries for the distribution
    :type: iterable

    :return: The uniform uncertainty value to apply to all of the measurements
        within the distribution
    :rtype: float
    """

    test = list(values) + []
    test.sort()

    deltas = []
    for i in range(len(test) - 1):
        deltas.append(abs(test[i] - test[i + 1]))

    return max(0.00001, 0.5 * float(np.median(deltas)))


def gaussian(x, measurement, max_sigma = 10.0):
    """
    A Gaussian kernel function that returns a probability density value at
    the given x position where the Gaussian is defined by the center (mean)
    and width (standard deviation)

    :param x: Position where the kernel should be evaluated to return the
        probability density value
    :type: float

    :param measurement: A ValueUncertainty instance that is used to define
        the parameters for the kernel
    :type: value.ValueUncertainty

    :param max_sigma: The maximum amount of sigmas from the measurement value
        where the kernel will calculate the probability density. Any values
        outside of this threshold will just return a value of 0.
    :type: float:

    :return: The probability for the measurement at the given x position
    :rtype: float
    """

    center = measurement.value
    width = measurement.uncertainty

    width = max(width, 1e-6)

    if x <= (center - max_sigma * width) or x >= (center + max_sigma * width):
        # Don't calculate values outside a "reasonable" 10 sigma range
        return 0.0

    coefficient = 1 / math.sqrt(2.0 * math.pi * width * width)
    exponent = -0.5 * ((float(x) - center) ** 2) / (width * width)

    return coefficient * math.exp(exponent)




from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from measurement_stats import value
from measurement_stats.density import kernels


def create(measurements, uncertainties=None, kernel=None):
    """
    Creates a distribution according to the specified arguments.

    If the measurements argument is an iterable of ValueUncertainty
    instances, they are used directly to create the distribution and any
    value for the uncertainties argument is ignored.

    If the measurements argument is an iterable of numeric types then
    they are converted into ValueUncertainties using the uncertainties
    argument.

    :param kernel: (optional) A function with the signature
        kernel(x, measurement) that returns the probability for the given
        measurement at the position value x. If no kernel is specified, the
        default Gaussian kernel will be used

    :param measurements: An iterable containing each of the measurements for
        the distribution. The iterable must be a homogeneous collection of
        either ValueUncertainty instances or a numeric type.

    :param uncertainties: (optional) If no uncertainties are specified, then
        an uncertainty is estimated based on the spacing between measurements.
        If uncertainties is a single numeric value, then that value is applied
        equally to all measurements. Otherwise, uncertainties should be an
        iterable with the same length as the measurements argument and contain
        numeric values of uncertainty for each measurement accordingly. This
        argument is ignored if the measurements are already ValueUncertainties.

    :return: A Distribution instance created based on the specified arguments
    :rtype: Distribution
    """

    if not measurements:
        return None

    measurements = list(measurements) + []

    if hasattr(measurements[0], 'raw'):
        # Already an iterable of ValueUncertainty instances
        return Distribution(kernel=kernel, measurements=measurements)

    if uncertainties is None:
        # If there is no uncertainty estimate one based on the spacing of the
        # points within the distribution and apply that to all of the
        # measurements
        unc = kernels.uncertainty_estimate(measurements)
        return Distribution(
            kernel=kernel,
            measurements=[
                value.ValueUncertainty(float(v), unc) for v in measurements
                ]
        )

    try:
        if len(measurements) != len(uncertainties):
            raise ValueError(
                'Measurements and uncertainties must be equal length iterables'
            )

        zipper = zip(measurements, uncertainties)
        return Distribution(
            kernel=kernel,
            measurements=[
                value.ValueUncertainty(float(v), float(u)) for v, u in zipper
                ]
        )
    except TypeError:
        # If the uncertainties are not a list then this will be raise and
        # we assume that uncertainties is just a number
        pass

    return Distribution(
        kernel=kernel,
        measurements=[
            value.ValueUncertainty(float(v), float(uncertainties))
            for v in measurements
            ]
    )


class Distribution(object):
    """
    Data structure representing a probability density distribution
    for a given set of measurements and a kernel function that defines
    how the probability for each measurement is distributed along the
    measurement (i.e. x) axis.

    Each measurement entry is assumed to account for a single measurement,
    so that it accounts for an equal amount of probability within the
    distribution.

    A measurement is a ValueUncertainty instance, which resolves using the
    kernel function, to its own probability distribution along the
    measurement axis (i.e. the x axis).
    """

    def __init__(self, measurements=None, kernel=None):
        self.kernel = kernel if kernel else kernels.gaussian
        self.measurements = measurements if measurements is not None else []

    def probability_at(self, x):
        """
        Returns the probability of the distribution at the given position
        on the measurement axis (x)

        :param x: The position along the measurement axis where the probability
            will be calculated
        :type: float

        :return: The probability (normalized to a maximum of 1.0) at the
            specified position on the measurement axis
        :rtype: float
        """

        return sum(
            [self.kernel(x, m) for m in self.measurements]
        ) / len(self.measurements)

    def probabilities_at(self, x_values):
        """
        The probabilities for the distribution at the given locations on
        the measurement (x) axis

        :param x_values: An iterable containing the values where the
            probability of the distribution should be calculated
        :type: iterable

        :return: A list containing the normalized probabilities at each of the
            specified position values (floats)
        :rtype: list
        """

        return [self.probability_at(x) for x in x_values]

    def minimum_value(self):
        """
        The lowest measurement value in the distribution

        :return: Minimum measurement value in the distribution
        :rtype: value.ValueUncertainty
        """

        if not len(self.measurements):
            return None

        item = self.measurements[0]
        for v in self.measurements[1:]:
            if v.value == item.value and v.uncertainty > item.uncertainty:
                item = v
            elif v.value < item.value:
                item = v

        return item

    def maximum_value(self):
        """
        The highest measurement value in the distribution

        :return: Maximum measurement value in the distribution
        :rtype: value.ValueUncertainty
        """

        if not len(self.measurements):
            return None

        point = self.measurements[0]
        for m in self.measurements[1:]:
            if m.value == point.value and m.uncertainty > point.uncertainty:
                point = m
            elif m.value > point.value:
                point = m

        return point

    def naked_measurement_values(self, raw=False):
        """
        Returns a list of naked numbers, i.e. without uncertainty values.

        :param raw: (optional) Specifies whether to return raw (not truncated)
            measurement values, or values that have been truncated by their
            uncertainties.
        :type: bool

        :return: A list of naked measurement values
        :rtype: list
        """

        return [(m.raw if raw else m.value) for m in self.measurements]

    def minimum_boundary(self, sigma_threshold):
        """
        The boundary is defined as the value on the measurement (x)
        axis where all measurements are above by at least the specified
        sigma deviation tolerance.

        For example, if we have a distribution with two measurements:
            * m1 = 12 +/- 2
            * m2 = 20 +/- 4
        and we specify sigma_threshold to be 10, then the boundary must be
        the minimum value of:
            * m1: 12 - 2 * 10 = -8
            * m2: 20 - 4 * 10 = -20
        or -20.

        :param sigma_threshold: The threshold number of sigmas deviations that
            all measurements must be above to define the boundary.
        :type: float

        :return: The boundary position at where all measurements reside above
            by at least the specified sigma threshold.
        :rtype: float
        """

        if not len(self.measurements):
            return 0.0
        return min([
                       (m.value - float(sigma_threshold) * m.uncertainty)
                       for m in self.measurements
                       ])

    def maximum_boundary(self, sigma_threshold):
        """
        The boundary is defined as the value on the measurement (x)
        axis where all measurements are below by at least the specified
        sigma deviation tolerance.

        For example, if we have a distribution with two measurements:
            * m1 = 12 +/- 2
            * m2 = 20 +/- 4
        and we specify sigma_threshold to be 10, then the boundary must be
        the maximum value of:
            * m1: 42 + 2 * 10 = 52
            * m2: 20 + 4 * 10 = 60
        or 60.

        :param sigma_threshold: The threshold number of sigmas deviations that
            all measurements must be below to define the boundary.
        :type: float

        :return: The boundary position at where all measurements reside below
            by at least the specified sigma threshold.
        :rtype: float
        """

        if not len(self.measurements):
            return 0.0
        return max([
                       (m.value + float(sigma_threshold) * m.uncertainty)
                       for m in self.measurements
                       ])




from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from measurement_stats.density import ops


def to_weighted_population(distribution_or_population, count=None):
    """

    :param distribution_or_population:
    :param count:
    :return:
    """

    if count is None:
        count = 4096

    if hasattr(distribution_or_population, 'measurements'):
        return ops.population(distribution_or_population, count)

    return distribution_or_population


def to_unweighted_population(distribution_or_values):
    """

    :param distribution_or_values:
    :return:
    """

    if hasattr(distribution_or_values, 'measurements'):
        return [x.value for x in distribution_or_values.measurements]

    return distribution_or_values



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys

if sys.version > '3':
    long = int


def sqrt_sum_of_squares(*args):
    """
    Computes the square-root of the sum of the squares for the combined list
    of numbers in the args list

    :param args:
    :return:
    """

    out = 0.0
    for arg in args:
        out += float(arg) * float(arg)

    return out ** 0.5


def is_number(value):
    """
    Determines whether or not the specified value is a numeric type, either
    integer or float.

    :param value:
    :return:
    """

    return isinstance(value, (int, long, float))


def linear_space(min_value=0, max_value=1.0, length=10, round_op=None):
    """
    Returns a list of linear-spaced values with min_value and max_value as
    the boundaries with the specified number (length) of entries. If
    roundToIntegers is True, each value will be rounded to the nearest
    integer.

    :param min_value:
    :param max_value:
    :param length:
    :param round_op:
    :return:
    """

    out = []
    value = min_value
    length = max(2, length)
    delta = (float(max_value) - float(min_value)) / float(length - 1.0)

    for index in range(length - 1):
        out.append(round_op(value) if round_op else value)
        value += delta

    out.append(round_op(max_value) if round_op else max_value)
    return out







from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import math

from measurement_stats import angle
from measurement_stats import ops
from measurement_stats import value


class Point2D(object):
    """A class for..."""

    def __init__(self, x=None, y=None):
        if x is None:
            x = value.ValueUncertainty()
        if y is None:
            y = value.ValueUncertainty()

        self.x = x
        self.y = y

    @property
    def length(self):
        """ The length of the vector

        :return:
        :rtype: value.ValueUncertainty
        """
        return self.distance_from(self.__class__())

    @property
    def nonzero(self):
        return self.x.value != 0.0 or self.x.value != 0.0

    def copy_from(self, source):
        """

        :param source:
        :return:
        """

        self.x.copy(source.x)
        self.y.copy(source.y)

    def clone(self):
        """

        :return:
        """
        return self.__class__(x=self.x.clone(), y=self.y.clone())

    def invert(self):
        """ Switches the sign of the x and y values so that x = -x and y = -y.
        """
        self.x.value = -self.x.value
        self.y.value = -self.y.value

    def rotate(self, rotation_angle, origin=None):
        """ Rotates the position value by the specified angle using a standard
            2D rotation matrix formulation. If an origin Position2D instance is
            not specified the rotation will occur around the origin. Also, if
            an origin is specified, the uncertainty in that origin value will
            be propagated through to the uncertainty of the rotated result.

        :param rotation_angle:
        :param origin: (optional)
        :return:
        """

        if origin is None:
            origin = self.__class__(
                value.ValueUncertainty(0, 0),
                value.ValueUncertainty(0, 0)
            )

        a = rotation_angle.radians
        x = self.x.raw - origin.x.raw
        y = self.y.raw - origin.y.raw

        self.x.update(
            x * math.cos(a) - y*math.sin(a) + origin.x.raw,
            ops.sqrt_sum_of_squares(self.x.uncertainty, origin.x.uncertainty)
        )
        self.y.update(
            y * math.cos(a) + x*math.sin(a) + origin.y.raw,
            ops.sqrt_sum_of_squares(self.y.uncertainty, origin.y.uncertainty)
        )

        return self

    def distance_from(self, point):
        """

        :param point:
        :return:
        """

        xDelta = self.x - point.x
        yDelta = self.y - point.y
        return (xDelta*xDelta + yDelta*yDelta) ** 0.5

    def serialize(self):
        """toDict doc..."""
        return dict(
            x=self.x.serialize(),
            y=self.y.serialize()
        )

    def normalize(self):
        """normalize doc..."""
        length = self.length
        if length == 0.0:
            return False

        self.x /= length
        self.y /= length

        return True

    def angle_between(self, point):
        """

        :param point:
        :return:
        """

        my_length = self.length
        pos_length = point.length
        numerator = self.x*point.x + self.y*point.y
        denominator = my_length * pos_length

        if value.equivalent(denominator.value, 0.0, 1e-6):
            return angle.Angle(radians=0.0, uncertainty=0.5*math.pi)

        result = numerator/denominator

        if value.equivalent(result.value, 1.0, 1e-5):
            return angle.Angle()

        try:
            if value.equivalent(result.value, -1.0, 1e-5):
                a = math.pi
            else:
                a = math.acos(result.raw)
        except Exception:
            print('[ERROR]: Unable to calculate angle between', result)
            return angle.Angle()

        if value.equivalent(a, math.pi, 1e-5):
            return angle.Angle(radians=a, uncertainty_degrees=180.0)

        try:
            aUnc = abs(1.0 / (1.0 - result.raw * result.raw) ** 0.5) * \
                   result.raw_uncertainty
        except Exception as err:
            print('[ERROR]: Unable to calculate angle between uncertainty',
                  result, a)
            return angle.Angle()

        return angle.Angle(radians=a, uncertainty=aUnc)

    def __pow__(self, power, modulo=None):
        try:
            return self.__class__(
                x=self.x ** power.x,
                y=self.y ** power.y
            )
        except Exception:
            return self.__class__(
                x=self.x ** power,
                y=self.y ** power
            )

    def __add__(self, other):
        try:
            return self.__class__(
                x=self.x + other.x,
                y=self.y + other.y
            )
        except Exception:
            return self.__class__(
                x=self.x + other,
                y=self.y + other
            )

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        try:
            return self.__class__(
                x=self.x - other.x,
                y=self.y - other.y
            )
        except Exception:
            return self.__class__(
                x=self.x - other,
                y=self.y - other
            )

    def __rsub__(self, other):
        try:
            return self.__class__(
                x=other.x - self.x,
                y=other.y - self.y
            )
        except Exception:
            return self.__class__(
                x=other - self.x,
                y=other - self.y
            )

    def __mul__(self, other):
        try:
            return self.__class__(x=self.x*other.x, y=self.y*other.y)
        except Exception:
            return self.__class__(x=self.x*other, y=self.y*other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        try:
            return self.__class__(x=self.x/other.x, y=self.y/other.y)
        except Exception:
            return self.__class__(x=self.x/other, y=self.y/other)

    def __rtruediv__(self, other):
        return self.__truediv__(other)

    def __div__(self, other):
        return self.__truediv__(other)

    def __rdiv__(self, other):
        return self.__rtruediv__(other)

    def __repr__(self):
        return self.__str__()

    def __unicode__(self):
        return '<{} {} {}>'.format(
            self.__class__.__name__,
            self.x.label,
            self.y.label
        )

    def __str__(self):
        return '{}'.format(self.__unicode__())







from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import random
from operator import itemgetter

import numpy as np

from measurement_stats import value
from measurement_stats.density.distribution import Distribution


def uniform_range(distribution, max_sigma, num_points=0, delta=0):
    """
    Creates a uniform range of values along the measurement (x) axis for
    the distribution within the boundaries created by the maximum sigma
    argument.

    :param distribution: The distribution for which the range should be created
    :type: refined_stats.density.Distribution

    :param max_sigma: Threshold sigma deviations that defines the range
        boundaries. The range will begin where all measurements are at least
        max_sigma above the value and end where all measurements are at
        least max_sigma below the value.
    :type: float

    :param num_points: (optional) The number of points to include in the range.
        If not specified the number of points will be assigned based either on
        the specified delta argument, or a default value of 1024.
    :type: int

    :param delta: (optional) Spacing between points within the range. If the
        number of points argument is specified (i.e. non-zero), this argument
        is ignored. If no number of points or delta are specified (both are 0)
        then a default delta will be assigned.
    :type: float

    :return: A list of uniformly spaced values that cover the distribution
        within the tolerance boundaries set by the max_sigma argument.
    :rtype: list
    """

    min_val = distribution.minimum_boundary(max_sigma)
    max_val = distribution.maximum_boundary(max_sigma)

    if not num_points:
        if delta:
            num_points = round((max_val - min_val) / delta)
        else:
            num_points = 1024

    return np.linspace(min_val, max_val, int(num_points))


def adaptive_range(distribution, max_sigma, max_delta=None):
    """
    Creates a dynamically assigned range of values along the measurement
    (x) axis for the distribution within the boundaries created by the
    maximum sigma argument.

    The values returned are not spaced equally. Instead they are spaced
    based on the gradient of the distribution, where larger gradients
    receive a higher density of values and smaller gradients less.

    :param distribution: The distribution for which the range should be created
    :type: refined_stats.density.Distribution

    :param max_sigma: Threshold sigma deviations that defines the range
        boundaries. The range will begin where all measurements are at least
        max_sigma above the value and end where all measurements are at
        least max_sigma below the value.
    :type: float

    :param max_delta: (optional) Maximum spacing between points within the
        range. If no maximum delta is assigned, one will be calculated
        automatically based on the structure of the distribution to prevent
        missing features.
    :type: float

    :return: A list of adaptively-spaced values that cover the distribution
        within the tolerance boundaries set by the max_sigma argument.
    :rtype: list
    """

    min_val = distribution.minimum_boundary(max_sigma)
    max_val = distribution.maximum_boundary(max_sigma)

    if max_delta is None:
        max_delta = 0.01 * abs(max_val - min_val)

    measurements = []
    for m in distribution.measurements:
        measurements.append({
            'min': m.value - 6.0*m.uncertainty,
            'max': m.value + 6.0*m.uncertainty,
            'value': m
        })

    measurements.sort(key=itemgetter('min'))

    out = [min_val]
    while out[-1] < max_val:
        delta = min(max_delta, max_val - out[-1])
        x_next = out[-1] + delta

        index = 0
        while len(measurements) and index < len(measurements):
            m = measurements[index]
            x = out[-1]
            x_next = x + delta

            if x_next <= m['min']:
                # BEFORE KERNEL: If the new x value is less than the lower
                # bound of the kernel value, use that delta value.
                break

            if m['max'] <= x:
                # AFTER KERNEL: If x is higher than the kernel upper
                # bound then skip to the next by removing this kernel from
                # the list so that it won't appear in future iterations.
                measurements.pop(0)
                continue

            if m['min'] <= x <= m['max']:
                # INSIDE KERNEL: If x is inside a kernel, use that kernel's
                # delta if it is smaller than the delta already set.
                delta = min(delta, 0.25*m['value'].uncertainty)

            elif x <= m['min'] and m['max'] <= x_next:
                # OVER KERNEL: If x and x+dx wrap around the kernel, use
                # a delta that puts the new x value at the lower edge of
                # the kernel.
                delta = min(delta, m['min'] - x)

            index += 1
            x_next = x + delta

        out.append(x_next)

    return out


def population(distribution, count=2048):
    """
    Creates a list of numerical values (no uncertainty) of the specified
    length that are representative of the distribution for use in less robust
    statistical operations.

    :param distribution:
        The distribution instance on which to create a population

    :param count: The number of numerical values to include in the returned
        population list.
    :type: int

    :return: A list of numerical values that approximate the the measurement
        probability density distribution
    :rtype: list
    """

    out = []
    x_min = distribution.minimum_boundary(10)
    x_max = distribution.maximum_boundary(10)

    x = x_min
    delta = (x_max - x_min) / 512.0
    total = float(len(distribution.measurements))

    while x <= x_max:
        n = int(round(count * delta * distribution.probability_at(x)))
        for i in range(n):
            out.append(random.uniform(
                x - 0.5 * delta,
                x + 0.5 * delta
            ))
        if x == x_max:
            break
        x = min(x_max, x + delta)
    return out


def overlap(distribution, comparison):
    """
    Returns the disparity in overlap between the two distributions.
    A value of 1.0 indicates that they overlap perfectly, while a value of
    0.0 indicates that there is no overlap between them. This approach can
    be useful in determining whether or not a distribution can be
    characterized by another one.

    :param distribution: A distribution for overlap comparison
    :type: refined_stats.density.Distribution

    :param comparison: A distribution for overlap comparison
    :type: refined_stats.density.Distribution

    :return: overlap on range [0, 1.0]
    :rtype: float
    """

    # Create a list of x values at critical points where the two distributions
    # should be evaluated
    x_values = list(set(
        adaptive_range(distribution, 10.0) +
        adaptive_range(comparison, 10.0)
    ))
    x_values.sort()

    out = 0.0
    my_value = distribution.probability_at(x_values[0])
    compare_value = comparison.probability_at(x_values[0])

    for i in range(len(x_values) - 1):
        x = x_values[i]
        x_next = x_values[i + 1]
        dx = x_next - x
        my_next_value = distribution.probability_at(x_next)
        compare_next_value = comparison.probability_at(x_next)

        my_area = dx * (
            min(my_value, my_next_value) +
            0.5 * abs(my_next_value - my_value)
        ) / float(len(distribution.measurements))

        compare_area = dx * (
            min(compare_value, compare_next_value) +
            0.5 * abs(compare_next_value - compare_value)
        )

        out += abs(my_area - compare_area)
        my_value = my_next_value
        compare_value = compare_next_value

    return 1.0 - 0.5*out


def weighted_median_average_deviation(distribution):
    """

    :param distribution:
    :return:
    """

    median = percentile(distribution, 0.5)

    # Create a list of the deviations, weighted by the uncertainty in the
    # measurement
    deviations = []
    for m in distribution.measurements:
        deviations.append(value.ValueUncertainty(
            abs(median - m.raw),
            m.raw_uncertainty
        ))

    dist = Distribution(
        measurements=deviations,
        kernel=distribution.kernel
    )

    return percentile(dist, 0.5)


def percentile(distribution_or_population, target=0.5, count=None):
    """
    Computes the position along the measurement axis where the distribution
    reaches the given target percentile_with_probability. The computation is
    carried out using a numerical integrator, which requires a tolerance where
    it can determine success and return

    :param distribution_or_population: The distribution for which the
        percentile should be calculated. Or a population list of values created
        by the distribution on which to calculate the percentile.

    :param target: The percentile_with_probability to find within the
        distribution.
    :type: float

    :param count: The number of points to use in populating the distribution
        for a standard percentile_with_probability calculation. Only matters if
        a distribution is specified, in which case the count is the size of the
        population create, which defaults to 4096. Ignored for populations as
        they already have  a count.
    :type: int
    """

    if count is None:
        count = 4096

    if hasattr(distribution_or_population, 'measurements'):
        pop = population(distribution_or_population, count=count)
    else:
        pop = distribution_or_population

    return np.percentile(pop, 100 * target)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys
import math
import random

if sys.version > '3':
    long = int


def order_of_magnitude(value):
    """
    Returns the order of magnitude of the most significant digit of the
    specified number. A value of zero signifies the ones digit, as would be
    the case in [Number]*10^[Order].

    :param value:
    :return:
    """

    x = abs(float(value))
    offset = 0 if x >= 1.0 else -1
    return int(math.log10(x) + offset)


def round_significant(value, digits):
    """

    :param value:
    :param digits:
    :return:
    """

    if value == 0.0:
        return 0

    value = float(value)
    d = math.ceil(math.log10(-value if value < 0  else value))
    power = digits - int(d)

    magnitude = math.pow(10, power)
    shifted = round(value*magnitude)
    return shifted / magnitude


def least_significant_order(value):
    """

    :param value:
    :return:
    """

    om = 0

    if isinstance(value, (int, long)) or long(value) == value:
        value = long(value)

        while om < 10000:
            om += 1
            magnitude = math.pow(10, -om)
            test = float(value) * magnitude
            if long(test) != test:
                return om - 1
        return 0

    while om < 10000:
        om -= 1
        magnitude = math.pow(10, -om)
        test = value * magnitude
        if equivalent(test, int(test)):
            return om
    return 0


def equivalent(a, b, epsilon=None, machine_epsilon_factor=100.0):
    """

    :param a:
    :param b:
    :param epsilon:
    :param machine_epsilon_factor:
    :return:
    """

    if epsilon is None:
        epsilon = machine_epsilon_factor * sys.float_info.epsilon
    return abs(float(a) - float(b)) < epsilon


def round_to_order(value, order, round_op=None, as_int=False):
    """

    :param value:
    :param order:
    :param round_op:
    :param as_int:
    :return:
    """

    if round_op is None:
        round_op = round

    if order == 0:
        value = round(float(value))
        return int(value) if as_int else value

    scale = math.pow(10, order)
    value = scale * round_op(float(value) / scale)
    return int(value) if as_int else value


class ValueUncertainty(object):
    """

    """

    def __init__(self, value =0.0, uncertainty =1.0, **kwargs):
        self.raw = float(value)
        self.raw_uncertainty = abs(float(uncertainty))

    @property
    def value(self):
        uncertainty = self.uncertainty
        order = least_significant_order(uncertainty)
        return round_to_order(self.raw, order)

    @value.setter
    def value(self, value):
        self.raw = float(value)

    @property
    def uncertainty(self):
        return round_significant(abs(self.raw_uncertainty), 1)

    @uncertainty.setter
    def uncertainty(self, value):
        self.raw_uncertainty = float(value)

    @property
    def html_label(self):
        return '%.15g &#177; %s' % (self.value, self.uncertainty)

    @property
    def label(self):
        return '%.15g +/- %s' % (self.value, self.uncertainty)

    @property
    def raw_label(self):
        return '%.15g +/- %.15g' % (
            round_significant(self.raw, 6),
            self.uncertainty
        )

    def from_dict(self, source):
        self.raw = source['raw']
        self.raw_uncertainty = source['raw_uncertainty']

    def to_dict(self):
        return dict(
            raw=self.raw,
            raw_uncertainty=self.raw_uncertainty
        )

    def serialize(self):
        return dict(
            value=self.value,
            uncertainty=self.uncertainty,
            raw=self.raw,
            raw_uncertainty=self.raw_uncertainty
        )

    def clone(self):
        """clone doc..."""
        return ValueUncertainty(
            value=self.raw,
            uncertainty=self.raw_uncertainty
        )

    def update(self, value = None, uncertainty = None):

        if value is not None:
            self.raw = value

        if uncertainty is not None:
            self.raw_uncertainty = uncertainty

    def freeze(self):
        """

        :return:
        :rtype: ValueUncertainty
        """
        self.raw = self.value
        self.raw_uncertainty = self.uncertainty
        return self

    @classmethod
    def create_random(
            cls, min_value =-1.0, max_value =1.0,
            min_uncertainty =0.1, max_uncertainty =2.0):
        """

        :param min_value:
        :param max_value:
        :param min_uncertainty:
        :param max_uncertainty:
        :return:
        """

        return ValueUncertainty(
            value=random.uniform(min_value, max_value),
            uncertainty=random.uniform(min_uncertainty, max_uncertainty) )

    def __pow__(self, power, modulo=None):
        if equivalent(self.raw, 0.0, 1e-5):
            return self.__class__(self.raw, self.raw_uncertainty)

        val = self.raw ** power
        unc = abs(val * float(power) * self.raw_uncertainty / self.raw)
        return self.__class__(val, unc)

    def __add__(self, other):
        try:
            val = self.raw + other.raw
            unc = math.sqrt(
                self.raw_uncertainty ** 2 +
                other.raw_uncertainty ** 2
            )
        except Exception:
            val = float(other) + self.raw
            unc = self.raw_uncertainty

        return self.__class__(val, unc)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        try:
            val = self.raw - other.raw
            unc = math.sqrt(
                self.raw_uncertainty ** 2 +
                other.raw_uncertainty ** 2
            )
        except Exception:
            val = self.raw - float(other)
            unc = self.raw_uncertainty

        return self.__class__(val, unc)

    def __rsub__(self, other):
        try:
            val = other.raw - self.raw
            unc = math.sqrt(
                self.raw_uncertainty ** 2 +
                other.raw_uncertainty ** 2
            )
        except Exception:
            val = float(other) - self.raw
            unc = self.raw_uncertainty

        return self.__class__(val, unc)

    def __mul__(self, other):
        try:
            val = self.raw * other.raw
            unc = abs(val)*math.sqrt(
                (self.raw_uncertainty / self.raw) ** 2 +
                (other.raw_uncertainty / other.raw) ** 2)
        except ZeroDivisionError:
            val = 0.0
            unc = math.sqrt(
                self.raw_uncertainty ** 2 +
                other.raw_uncertainty ** 2
            )
        except Exception:
            val = float(other) * self.raw
            unc = abs(float(other) * self.raw_uncertainty)

        return self.__class__(val, unc)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        try:
            val = self.raw/other.raw
            unc = abs(val) * math.sqrt(
                (self.raw_uncertainty / self.raw) ** 2 +
                (other.raw_uncertainty / other.raw) ** 2)
        except ZeroDivisionError:
            if equivalent(self.raw, 0.0, 1e-6):
                val = 0.0
                unc = math.sqrt(
                    self.raw_uncertainty ** 2 +
                    other.raw_uncertainty ** 2
                )
            else:
                raise
        except Exception:
            val = self.raw / float(other)
            unc = abs(self.raw_uncertainty / float(other))

        return self.__class__(val, unc)

    def __rtruediv__(self, other):
        return self.__truediv__(other)

    def __div__(self, other):
        return self.__truediv__(other)

    def __rdiv__(self, other):
        return self.__rtruediv__(other)

    def __repr__(self):
        return self.__str__()

    def __unicode__(self):
        """__unicode__ doc..."""
        return '<%s %s>' % (self.__class__.__name__, self.label)

    def __str__(self):
        return '<%s %s>' % (self.__class__.__name__, self.label)




from setuptools import setup
from setuptools import find_packages


def read_me():
    with open('README.rst') as f:
        return f.read()

setup(
    name='measurement_stats',
    version='0.2.1',
    description=
            'Measurement statistics with uncertainties and error propagation',
    long_description=read_me(),
    url='https://github.com/sernst/Measurement_Statistics',
    author='Scott Ernst',
    author_email='swernst@gmail.com',
    license='MIT',
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',

        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Physics'
    ],
    install_requires=[
        'pandas',
        'numpy',
        'six'
    ],
    test_suite='nose.collector',
    tests_require=['nose'],
    keywords='measurements statistics error propagation',
)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import math

from measurement_stats import value


class Angle(object):
    """A type for representing angular measurements with uncertainties"""

    def __init__(self, **kwargs):
        self._angle = 0.0
        self._unc = 1.0

        if 'uncertainty' in kwargs:
            self.uncertainty = kwargs.get('uncertainty')
        elif 'uncertainty_degrees' in kwargs:
            self.uncertainty_degrees = kwargs.get('uncertainty_degrees')

        if 'degrees' in kwargs:
            self.degrees = kwargs.get('degrees')
        elif 'radians' in kwargs:
            self.radians = kwargs.get('radians')

    @property
    def value(self):
        return value.ValueUncertainty(self.radians, self.uncertainty)

    @property
    def value_degrees(self):
        return value.ValueUncertainty(self.degrees, self.uncertainty_degrees)

    @property
    def uncertainty(self):
        return self._unc

    @uncertainty.setter
    def uncertainty(self, v):
        self._unc = v

    @property
    def uncertainty_degrees(self):
        return math.degrees(self._unc)

    @uncertainty_degrees.setter
    def uncertainty_degrees(self, v):
        self._unc = math.radians(v)

    @property
    def radians(self):
        return self._angle

    @radians.setter
    def radians(self, v):
        self._angle = float(v)

    @property
    def degrees(self):
        return math.degrees(self._angle)

    @degrees.setter
    def degrees(self, v):
        self._angle = math.radians(float(v))

    @property
    def pretty_print(self):
        return value.round_significant(self.degrees, 3)

    def clone(self):
        """clone doc..."""
        return self.__class__(radians=self._angle, uncertainty=self._unc)

    def constrain_to_revolution(self):
        """
        Constrains the angle to within the bounds [0, 360] by removing
        revolutions.
        """

        radians = self.radians
        while radians < 0:
            radians += 2.0*math.pi
        degrees = math.degrees(radians) % 360.0
        self.radians = math.radians(degrees)
        return self

    def difference_between(self, angle):
        """
        Returns a new Angle instance that is the smallest difference between
        this angle the one specified in the arguments.

        :param angle:
        :return:
        :rtype: Angle
        """

        a = self.clone().constrain_to_revolution()
        b = angle.clone().constrain_to_revolution()

        result = a - b
        return Angle(
            degrees=((result.degrees + 180.0) % 360.0) - 180.0,
            uncertainty=result.uncertainty
        )

    def __pow__(self, power, modulo=None):
        v = self.value ** power
        return self.__class__(
            radians=v.raw,
            uncertainty=v.raw_uncertainty
        )

    def __add__(self, other):
        v = self.value + other.value
        return self.__class__(
            radians=v.raw,
            uncertainty=v.raw_uncertainty
        )

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        v = self.value - other.value
        return self.__class__(
            radians=v.raw,
            uncertainty=v.raw_uncertainty
        )

    def __rsub__(self, other):
        return self.__sub__(other)

    def __mul__(self, other):
        v = self.value * other.value
        return self.__class__(
            radians=v.raw,
            uncertainty=v.raw_uncertainty
        )

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        v = self.value / other.value
        return self.__class__(
            radians=v.raw,
            uncertainty=v.raw_uncertainty
        )

    def __rtruediv__(self, other):
        return self.__truediv__(other)

    def __div__(self, other):
        return self.__truediv__(other)

    def __rdiv__(self, other):
        return self.__rtruediv__(other)

    def __repr__(self):
        return self.__str__()

    def __unicode__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.value.label)

    def __str__(self):
        return '{}'.format(self.__unicode__())



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import math
import numpy as np

from measurement_stats import value


def unweighted(*values):
    """

    :param values:
    :return:
    """

    if not values:
        return value.ValueUncertainty()

    if isinstance(values[0], (list, tuple)):
        values = values[0]
        if not values:
            return value.ValueUncertainty()

    values = [v.value for v in values]
    return value.ValueUncertainty(
            value=float(np.mean(values)),
            uncertainty=float(np.std(values))
    )


def weighted(*values):
    """
    Calculates the uncertainty weighted average of the provided values,
    where each value is a ValueUncertainty instance. For mathematical
    formulation of the weighted average see "An Introduction to Error
    Analysis, 2nd Edition" by John R. Taylor, Chapter 7.2.

    :param values:
    :return:
    """

    if not values:
        return value.ValueUncertainty()

    if isinstance(values[0], (list, tuple)):
        values = values[0]
        if not values:
            return value.ValueUncertainty()

    wxs = 0.0
    ws  = 0.0
    for v in values:
        w = 1.0/(v.uncertainty*v.uncertainty)
        wxs += w*v.value
        ws  += w

    ave = wxs/ws
    unc =  1.0/math.sqrt(ws)

    return value.ValueUncertainty(value=ave, uncertainty=unc)


def weighted_mean_and_deviation(*values):
    """
    Returns the mean and standard deviation of a weighted set of values.
    For further info see:
        http://stats.stackexchange.com/questions/6534/
            how-do-i-calculate-a-weighted-standard-deviation-in-excel

    :param values:
    :return:
    """

    if not values:
        return value.ValueUncertainty()

    if isinstance(values[0], (list, tuple)):
        values = values[0]
        if not values:
            return value.ValueUncertainty()

    if len(values) == 1:
        return values[0].clone()

    wxs = 0.0
    ws  = 0.0
    weights = []

    for v in values:
        w = 1.0/(v.uncertainty*v.uncertainty)
        weights.append(w)
        wxs += w*v.value
        ws  += w

    ave = wxs/ws
    dev = 0.0
    N = len(values)
    for i in range(N):
        dev += weights[i]*(values[i].value - ave)**2

    denom = ws*(N - 1.0)/N
    dev = math.sqrt(dev/denom)

    return value.ValueUncertainty(value=ave, uncertainty=dev)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np

from measurement_stats.density.boundaries import support


def unweighted_tukey(distribution_or_values):
    """
    Returns the unweighted Tukey box-whisker boundaries for the given
    distribution, including the median. As the returned boundaries are
    unweighted, they ignore uncertainties in the measurements.

    :param distribution_or_values:
        The distribution for which to calculate the boundaries
    :return:
        A tuple containing 5 values:
            * minimum (whisker boundary)
            * lower quartile (box boundary)
            * median
            * upper quartile (box boundary)
            * maximum (whisker boundary)
    :rtype: tuple
    """

    values = support.to_unweighted_population(distribution_or_values)

    median = np.percentile(values, 50)
    lower_quartile = np.percentile(values, 25)
    upper_quartile = np.percentile(values, 75)

    inter_quartile_range = upper_quartile - lower_quartile
    tukey_range = 1.5 * inter_quartile_range

    values.sort()
    bottom = None
    top = None
    for index in range(len(values)):
        v = values[index]
        if bottom is None and v >= (median - tukey_range):
            bottom = v

        v = values[len(values) - index - 1]
        if top is None and v <= (median + tukey_range):
            top = v

        if top is not None and bottom is not None:
            break

    return bottom, lower_quartile, median, upper_quartile, top


def weighted_tukey(distribution_or_population, count=None):
    """
    Returns the weighted Tukey box-whisker boundaries for the given
    distribution, including the mean. These boundaries are based on the
    weighted density distribution and take into account the uncertainties
    in the measurements.

    :param distribution_or_population:
        The distribution for which to calculate the boundaries
    :param count:
        The number of values to use when creating the weighted probability
        population used to define the whiskers
    :return:
        A tuple containing 5 values:
        * minimum (whisker boundary)
        * lower quartile (box boundary)
        * median
        * upper quartile (box boundary)
        * maximum (whisker boundary)
    :rtype: tuple
    """

    return unweighted_tukey(support.to_weighted_population(
        distribution_or_population, count
    ))


def unweighted_nine(distribution_or_values):
    """
    Returns the unweighted "Nines" box-whisker boundaries for the given
    distribution, including the median, where the whiskers are defined by the
    9% and 91% percentiles.

    :param distribution_or_values:
        The distribution for which to calculate the boundaries
    :return:
        A tuple containing 5 values:
            * minimum (whisker boundary)
            * lower quartile (box boundary)
            * median
            * upper quartile (box boundary)
            * maximum (whisker boundary)
    :rtype: tuple
    """

    values = support.to_unweighted_population(distribution_or_values)

    return (
        np.percentile(values, 9),
        np.percentile(values, 25),
        np.percentile(values, 50),
        np.percentile(values, 75),
        np.percentile(values, 91)
    )


def weighted_nine(distribution_or_population, count=None):
    """
    Returns the weighted "Nines" box-whisker boundaries for the given
    distribution, including the median, where the whiskers are defined by the
    9% and 91% percentiles.

    :param distribution_or_population:
        The distribution for which to calculate the boundaries
    :param count:
        The number of values to use when creating the weighted probability
        population
    :return:
        A tuple containing 5 values:
            * minimum (whisker boundary)
            * lower quartile (box boundary)
            * median
            * upper quartile (box boundary)
            * maximum (whisker boundary)
    :rtype: tuple
    """

    return unweighted_nine(
        support.to_weighted_population(distribution_or_population, count)
    )


def unweighted_two(distribution_or_values):
    """
    Returns the unweighted "Twos" box-whisker boundaries for the given
    distribution, including the median, where the whiskers are defined by the
    2% and 98% percentiles.

    :param distribution_or_values:
        The distribution for which to calculate the boundaries
    :return:
        A tuple containing 5 values:
            * minimum (whisker boundary)
            * lower quartile (box boundary)
            * median
            * upper quartile (box boundary)
            * maximum (whisker boundary)
    :rtype: tuple
    """

    values = support.to_unweighted_population(distribution_or_values)

    return (
        np.percentile(values, 2),
        np.percentile(values, 25),
        np.percentile(values, 50),
        np.percentile(values, 75),
        np.percentile(values, 98)
    )


def weighted_two(distribution_or_population, count=None):
    """
    Returns the weighted "Twos" box-whisker boundaries for the given
    distribution, including the median, where the whiskers are defined by the
    2% and 98% percentiles.

    :param distribution_or_population:
        The distribution for which to calculate the boundaries
    :param count:
        The number of values to use when creating the weighted probability
        population
    :return:
        A tuple containing 5 values:
            * minimum (whisker boundary)
            * lower quartile (box boundary)
            * median
            * upper quartile (box boundary)
            * maximum (whisker boundary)
    :rtype: tuple
    """

    return unweighted_two(
        support.to_weighted_population(distribution_or_population, count)
    )



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from measurement_stats.density import kernels
from measurement_stats.density import ops
from measurement_stats.density import boundaries
from measurement_stats.density.distribution import create as create_distribution
from measurement_stats.density.distribution import Distribution




from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import math


def unzip(values, raw=False):
    """

    :param values:

    :param raw: (optional) If True, the values and uncertainty lists will be
        raw instead the standard truncated formats specified by error
        propagation rules
    :return: tuple containing a list of values and a list of uncertainties
    """

    vals = []
    uncs = []

    if isinstance(values[0], (list, tuple)):
        values = values[0]

    for v in values:
        vals.append(v.raw if raw else v.value)
        uncs.append(v.raw_uncertainty if raw else v.uncertainty)

    return vals, uncs


def deviations(expected, values):
    """

    :param expected:
    :param values:
    :return:
    """

    out = []

    if hasattr(expected, 'raw_uncertainty'):
        for v in values:
            err = math.sqrt(
                v.raw_uncertainty ** 2 +
                expected.raw_uncertainty ** 2
            )
            out.append(abs(v.raw - expected.raw) / err)
    else:
        for v in values:
            out.append(abs(v.value - expected) / v.uncertainty)

    return out
